const char inp_split_doc[]="
Construct a new Input object with the input source given as argument.
This source is removed from the current input and transfered to the newly
created Input object.

Parameters: (i) : the input origin to split.
Return:     O : a new input object.
";
static PyObject* inp_split(GIIInputObject * self, PyObject* args) {
	gii_input_t inp;
	GIIInputObject *ret;
	long orig;
	int r;
	
	if(self->inp==NULL) {
		PyErr_SetString(PyExc_TypeError,"Input invalid");
		return NULL;
	}
	
	if(!PyArg_ParseTuple(args, "i", &orig)) {
		PyErr_SetString(PyExc_TypeError,"origin required");
		return NULL;
	}

	r=giiSplitInputs(self->inp,&inp,orig,0);
	
	if(r==GGI_ENOTFOUND) {
		PyErr_SetString(PyExc_ValueError,"device not found");
		return NULL;
	}
	ret=PyObject_New(GIIInputObject, &input_type);
	if(!ret) return NULL;
	ret->inp=inp;
	return (PyObject*)ret;
}


const char inp_join_doc[]="
Merge the given input object in this one. The second input becomes invalid and
should not be used afterwards. It is advised to explicitely destroy it with
the 'del' builtin function.

Parameters: (O) : a valid input.
Return:     None
";
static PyObject* inp_join(GIIInputObject * self, PyObject* args) {
	PyObject * inp;
	
	if(self->inp==NULL) {
		PyErr_SetString(PyExc_TypeError,"Input invalid");
		return NULL;
	}

	if(!PyArg_ParseTuple(args, "O", &inp)) {
		PyErr_SetString(PyExc_TypeError,"Input required");
		return NULL;
	}
	
	if(!GIIInput_Check(inp)) {
		PyErr_SetString(PyExc_TypeError,"Input required");
		return NULL;
	}
	
	if(((GIIInputObject*)inp)->inp==NULL) {
		PyErr_SetString(PyExc_TypeError,"valid Input required");
		return NULL;
	}

	self->inp=giiJoinInputs(self->inp,((GIIInputObject*)inp)->inp);
	((GIIInputObject*)inp)->inp=NULL; /* invalidate the object */
	Py_INCREF(Py_None);
	return Py_None;
}

/*------------- Event management  ------------------------*/

const char inp_eventread_doc[]="
Return the next event available on the queue.

Parameters: ()|(i) : the event mask (emAll if not set).
Return:     E: the next event of the given mask, if any
";
PyObject* inp_eventread(GIIInputObject *self, PyObject *args) {
	int mask;
	gii_event e;
	
	if(self->inp==NULL) {
		PyErr_SetString(PyExc_TypeError,"Input invalid");
		return NULL;
	}
	
	mask=emAll; /* by default */
	if(!PyArg_ParseTuple(args, "|i", &mask)) {
		return NULL;
	}

	if(giiEventsQueued(self->inp,mask)) {
		giiEventRead(self->inp, &e, mask);
		return newEventObjectFromEvent(e);
	}
	Py_INCREF(Py_None);
	return Py_None;
}


const char inp_eventblock_doc[]="
Block until an event in the given mask is available.

Parameters: ()|(i) : the event mask (emAll if not set).
Return:     The mask of available event.
";
PyObject* inp_eventblock(GIIInputObject *self, PyObject *args) {
	int mask;
	
	if(self->inp==NULL) {
		PyErr_SetString(PyExc_TypeError,"Input invalid");
		return NULL;
	}
	
	mask=emAll; /* by default */
	if(!PyArg_ParseTuple(args, "|i", &mask)) {
		return NULL;
	}
	
	return PyInt_FromLong(giiEventPoll(self->inp, mask, NULL));
}

const char inp_eventpoll_doc[]="
Wait for an event in the given mask to be available on the queue.

Parameters: () or one or both (in this order) of 
              - i : the mask (defaults to emAll)
              - [ sec, usec ] : timeout (ints, defaults to 0)
Return:     None : no event occured before timeout
            -i : the mask of available events, if no timeout was specified
            - [ i,[ sec, usec ] ] : same as above but with the remaining time.
";
PyObject* inp_eventpoll(GIIInputObject *self, PyObject*args) {
	struct timeval t;
	int mask;
	int hastime;

	if(self->inp==NULL) {
		PyErr_SetString(PyExc_TypeError,"Input invalid");
		return NULL;
	}
	mask=emAll;
	t.tv_usec=0;
	t.tv_sec=0;
	hastime=0;

	switch(PyTuple_Size(args)) {
	case 0:
		break;
	case 1:
		if(PyArg_ParseTuple(args, "i", &mask))
			break;		
		PyErr_Clear();
		if(PyArg_ParseTuple(args, "(ll)", &t.tv_sec,&t.tv_usec)) {
			hastime=1;
			break;
		}
		PyErr_Clear();
		PyErr_SetString(PyExc_TypeError, "Need a event mask and/or a timeout");
		return NULL;
	case 2:
		if(PyArg_ParseTuple(args, "i(ll)",&mask, &t.tv_sec, &t.tv_usec)) {
			hastime=1;
			break;
		}
		return NULL;
	default:
		PyErr_SetString(PyExc_TypeError, "Need at most an event mask and/or a timeout");
		return NULL;
	}
	
	mask=giiEventPoll(self->inp,mask,&t);

	if (mask==0) {
		Py_INCREF(Py_None);
		return Py_None;
	}
	if(hastime)
		return Py_BuildValue("[i[ll]]", mask, t.tv_sec, t.tv_usec);
	return PyInt_FromLong(mask);
}

const char inp_eventsqueued_doc[]="
Return the number of events waiting in the queue.

Parameters: ()|(i) : the event mask i (emAll if not set)
Return:   sure.
";
PyObject* inp_eventsqueued(GIIInputObject *self, PyObject*args) {
	int mask;
	
	if(self->inp==NULL) {
		PyErr_SetString(PyExc_TypeError,"not a valid input");
		return NULL;
	}
	
	mask=emAll; /* by default */
	if(!PyArg_ParseTuple(args, "|i", &mask)) {
		return NULL;
	}
	return PyInt_FromLong(giiEventsQueued(self->inp, mask));
}

const char inp_eventsend_doc[]="
Inject an event given as argument to the input source.
The event must be constructed before.

Parameters: (O) : the event
Return:     None
";
PyObject* inp_eventsend(GIIInputObject *self, PyObject* args) {
	PyObject * e;

	if(self->inp==NULL) {
		PyErr_SetString(PyExc_TypeError,"Input invalid");
		return NULL;
	}

	if(!PyArg_ParseTuple(args, "O", &e)) {
		PyErr_SetString(PyExc_TypeError, "Event required");
		return NULL;
	}
	
	if(!GIIEvent_Check(e)) {
		PyErr_SetString(PyExc_TypeError,"Event required");
		return NULL;
	}
	
	if(giiEventSend(self->inp, &(((GIIEventObject*)e)->ev))<0) {
		PyErr_SetString(PyExc_SystemError,"giiEventSend failed");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}


enum {
	PYGII_METH_JOIN = 0,
	PYGII_METH_SPLIT,
	PYGII_METH_READ,
	PYGII_METH_POLL,
	PYGII_METH_BLOCK,
	PYGII_METH_QUEUED,
	PYGII_METH_SEND
};

static struct PyMethodDef inp_methods[] = {
	/* ---------- Input management ----------------*/
	{ "join",    (PyCFunction)inp_join,    METH_VARARGS, (char*)inp_join_doc  },
	{ "split",   (PyCFunction)inp_split,   METH_VARARGS, (char*)inp_split_doc },
	/*------------- Events ------------------------*/
	{ "read", (PyCFunction)inp_eventread, METH_VARARGS, (char*)inp_eventread_doc },
	{ "poll", (PyCFunction)inp_eventpoll, METH_VARARGS, (char*)inp_eventpoll_doc },
	{ "wait", (PyCFunction)inp_eventblock, METH_VARARGS, (char*)inp_eventblock_doc },
	{ "queued", (PyCFunction)inp_eventsqueued, METH_VARARGS, (char*)inp_eventsqueued_doc},
	{ "send",    (PyCFunction)inp_eventsend,    METH_VARARGS, (char*)inp_eventsend_doc},
	{NULL, NULL}
};

/**
 ** Destructor
 **/
static void inp_dealloc(GIIInputObject *self) {
	if(self->inp)
		giiClose(self->inp);
	PyObject_Del(self);
}


static int inp_setattr(GIIInputObject*self, char*name, PyObject * value) {
	int mask;
	switch(name[0]) {
	case 'm':
		if(!strcmp(name,"mask")) {
			if(pygii_get_int(value, &mask)<0)
				return -1;
			giiSetEventMask(self->inp, mask);
			return 0;
		}
		break;
	}
	PyErr_SetString(PyExc_AttributeError, name);
	return -1;
}

static PyObject* inp_getattr(GIIInputObject*self, char*name) {
	switch(name[0]) {
	case 'b':
		break;
	case 'd':
		if(!strcmp(name,"devices")) {
			if(self->inp==NULL) {
				PyErr_SetString(PyExc_TypeError,"Input invalid");
				return NULL;
			}
			return pygii_device_info(self->inp);
		}
		break;
	case 'j':
		if(!strcmp(name,"join"))
			return PyCFunction_New(&inp_methods[PYGII_METH_JOIN],(PyObject*)self);
		break;
	case 'm':
		if(!strcmp(name,"mask"))
			return PyInt_FromLong(giiGetEventMask(self->inp));
		break;
	case 'p':
		if(!strcmp(name,"poll"))
			return PyCFunction_New(&inp_methods[PYGII_METH_POLL],(PyObject*)self);
		break;
	case 'q':
		if(!strcmp(name,"queued"))
			return PyCFunction_New(&inp_methods[PYGII_METH_QUEUED],(PyObject*)self);
		break;
	case 'r':
		if(!strcmp(name,"read"))
			return PyCFunction_New(&inp_methods[PYGII_METH_READ],(PyObject*)self);
		break;
	case 's':
		if(!strcmp(name,"send"))
			return PyCFunction_New(&inp_methods[PYGII_METH_SEND],(PyObject*)self);
		if(!strcmp(name,"split"))
			return PyCFunction_New(&inp_methods[PYGII_METH_SPLIT],(PyObject*)self);
		break;
	case 'w':
		if(!strcmp(name,"wait"))
			return PyCFunction_New(&inp_methods[PYGII_METH_BLOCK],(PyObject*)self);
		break;
	}
	PyErr_SetString(PyExc_AttributeError, name);
	return NULL;
}

/**
 ** Input object type.
 ** Other modules will get the address from CObject.
 **/
static PyTypeObject input_type = {
	PyObject_HEAD_INIT(NULL)
	0,
	"GII Input",
	sizeof(GIIInputObject),
	0,
	(destructor)inp_dealloc,
	(printfunc)0,
	(getattrfunc)inp_getattr,
	(setattrfunc)inp_setattr,
	(cmpfunc)0,
	(reprfunc)0,
	0,
	0,
	0,
	(hashfunc)0,
	(ternaryfunc)0,
	(reprfunc)0,
	
	0L,0L,0L,0L,
	"no docs."
};

static PyObject* newInputObject(PyObject*targets) {
	GIIInputObject *self;
	char *args[10];
	PyObject *t;
	int i,size;
	
	self=PyObject_New(GIIInputObject, &input_type);
	if(!self) return NULL;
	
	size=PyTuple_Size(targets);
	if(size>10) {
		PyErr_SetString(PyExc_TypeError,"can only manage 10 arguments.");
		return NULL;
	}
	
	for(i=0;i<size; i++) {
		t=PyTuple_GetItem(targets, i);
		if(!PyString_Check(t)) {
			PyErr_SetString(PyExc_TypeError, "strings only.");
			return NULL;
		}
		args[i]=PyString_AsString(t);
	}
	
	self->inp=_giiOpenWithArgs[size](args);
	if(self->inp==NULL) {
		PyErr_SetString(PyExc_SystemError, "giiOpen failed");
		return NULL;
	}
	
	return (PyObject*)self;
}

